/*
 * All this code come from V4L2 API.
 *
 * It is under GPL Licence.
 *
 * patched from Luca Amore and Marco Comparato
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <linux/compiler.h>
#include <linux/videodev.h>


static struct v4l2_queryctrl queryctrl;
static struct v4l2_querymenu querymenu;

extern int errno ;

/*
 * V4L2 supported controls by device
 */

struct supported_control {
	__u32	id ;
	int		current_value ;
} ;

 
struct supported_v4l2_queryctrl {
	struct supported_control *s_control ; 
	__u32	nb_controls ;
} ;


static struct supported_v4l2_queryctrl *supported_values = NULL ;

static void init_supported_controls( __u32 control_id, __s32 default_value )
{
	if( supported_values == NULL ) {
		supported_values = malloc( sizeof( struct supported_v4l2_queryctrl ) );
		if( supported_values == NULL ) {
			perror("Lack of memory" );
			exit(1);
		} else {
			supported_values->s_control = NULL ;
			supported_values->nb_controls = 0 ;
		}
	}

	supported_values->s_control = (struct supported_control*) realloc( 
						supported_values->s_control, 
						(supported_values->nb_controls+1)*sizeof(struct supported_control));


	if( supported_values->s_control == NULL ) {
		perror("Lack of memory on s_control allocation" );
		exit(1);
	} else {

#ifdef DEBUG
		printf("We add: %d %d\n", control_id, default_value );
#endif
		supported_values->s_control[ supported_values->nb_controls ].id = control_id ;
		supported_values->s_control[ supported_values->nb_controls ].current_value = default_value ;
		( supported_values->nb_controls )++ ;
	}
	
}

static void set_current_value_for_control( __u32 control_id, int value )
{
	unsigned i;
	
	for( i=0 ; i < supported_values->nb_controls ; i++ ) {
		if( supported_values->s_control[i].id == control_id ) {
			supported_values->s_control[i].current_value = value  ;
		}
	}
}


static int	get_current_value_for_control( __u32 control_id ) 
{
	unsigned i;

	for( i=0 ; i<supported_values->nb_controls ; i++ ) {
		if( supported_values->s_control[i].id == control_id ) {
			return supported_values->s_control[i].current_value ;
		}
	}
	return -1 ;
}

static void enumerate_menu ( int fd )
{
	printf ("  Menu items:\n");

	memset (&querymenu, 0, sizeof (querymenu));
	
	querymenu.id = queryctrl.id;

	for (querymenu.index = queryctrl.minimum;
       querymenu.index <= queryctrl.maximum; querymenu.index++) {
		if (0 == ioctl (fd, VIDIOC_QUERYMENU, &querymenu)) {
			printf ("  *  %s\n", querymenu.name);
		} else {
			perror ("VIDIOC_QUERYMENU");
			exit( -1 );
		}
	}
}

void enumerate_controls_on_device( int fd )
{
	memset (&queryctrl, 0, sizeof (queryctrl));
	printf("V4L2 Standard controls supported by device : \n" );
	
	for( queryctrl.id = V4L2_CID_BASE; 
					queryctrl.id < V4L2_CID_LASTP1; queryctrl.id++) {
		if( 0 == ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl)  ) {
			if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
				continue;
			printf ("=> Control %s\n", queryctrl.name);
			init_supported_controls( queryctrl.id, queryctrl.default_value );
			if (queryctrl.type == V4L2_CTRL_TYPE_MENU) {
				enumerate_menu ( fd );
			}
      	} else {
			if (errno == EINVAL)
			  continue;
			perror ("VIDIOC_QUERYCTRL");
			exit (EXIT_FAILURE);
		}
	}

	printf("Device designer query controls : \n") ;
	for (queryctrl.id = V4L2_CID_PRIVATE_BASE;; queryctrl.id++) {
		if ( 0 == ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl) ) {
			if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
				continue;
			}
			printf ("=> Control %s\n", queryctrl.name);
			if (queryctrl.type == V4L2_CTRL_TYPE_MENU) {
				enumerate_menu( fd );
			}
		} else {
			if (errno == EINVAL)
				break;

			perror ("VIDIOC_QUERYCTRL");
			exit (EXIT_FAILURE);
      	}
	}
}

/*
 * This code is from :
 * See : http://v4l2spec.bytesex.org/spec/control.html
 * Under GPL
 */
int change_ctrl( int fd , __u32 ctrl_id, int down_up ) 
{

	struct v4l2_queryctrl 	queryctrl;
	struct v4l2_control 	control;

	memset (&queryctrl, 0, sizeof (queryctrl));
	queryctrl.id = ctrl_id ;

	if( -1 == ioctl (fd, VIDIOC_QUERYCTRL, &queryctrl)) {
		if (errno != EINVAL) {
			perror ("VIDIOC_QUERYCTRL");
			exit (EXIT_FAILURE);
		} else {
			printf ("This control is not supported by your device.\n");
		}
	} else if( queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
		printf ("This control is not supported by your device.\n");
	} else {
		memset (&control, 0, sizeof (control));
		control.id = ctrl_id ;
		if( down_up > 0 ) {
			if( get_current_value_for_control(queryctrl.id) + queryctrl.step < queryctrl.maximum ) {
				control.value =  get_current_value_for_control(queryctrl.id) + queryctrl.step ;
				set_current_value_for_control( queryctrl.id, control.value );
				if( -1 == ioctl (fd, VIDIOC_S_CTRL, &control) ) {
					perror ("VIDIOC_S_CTRL");
					exit (EXIT_FAILURE);
				}
			} else {
				printf("This control have reached maximun value.\n" );
			}
		} else {
			if( get_current_value_for_control(queryctrl.id) - queryctrl.step < queryctrl.minimum ) {
				printf("This control have reached minimun.\n" );
			} else {
				control.value =  get_current_value_for_control(queryctrl.id) - queryctrl.step ;
				set_current_value_for_control( queryctrl.id, control.value );
				if( -1 == ioctl (fd, VIDIOC_S_CTRL, &control) ) {
					perror ("VIDIOC_S_CTRL");
					exit (EXIT_FAILURE);
				}
			}
		}
	}
	return 0 ;
}
